import jwt
import requests
import string
from datetime import datetime, timezone
from logging import Logger
from pypomes_core import str_random
from requests import Response
from threading import Lock
from typing import Any

from .jwt_constants import (
    JWT_DEFAULT_ALGORITHM, JWT_ENCODING_KEY,
    JWT_ROTATE_TOKENS, JWT_DB_ENGINE, JWT_DB_TABLE
)


class JwtData:
    """
    Shared JWT data for security token access.

    Instance variables:
        - access_lock: lock for safe multi-threading access
        - access_data: dictionary holding the JWT token data, organized by account id:
         {
           <account-id>: {
             "reference-url":              # the reference URL
             "remote-provider": <bool>,    # whether the JWT provider is a remote server
             "request-timeout": <int>,     # in seconds - defaults to no timeout
             "access-max-age": <int>,      # in seconds - defaults to JWT_ACCESS_MAX_AGE
             "refresh-max-age": <int>,     # in seconds - defaults to JWT_REFRESH_MAX_AGE
             "grace-interval": <int>       # time to wait for token to be valid, in seconds
             "token-audience": <string>    # the audience the token is intended for
             "token_nonce": <string>       # value used to associate a client session with a token
             "claims": {
               "birthdate": <string>,    # subject's birth date
               "email": <string>,        # subject's email
               "gender": <string>,       # subject's gender
               "name": <string>,         # subject's name
               "roles": <List[str]>,     # subject roles
               "nonce": <string>,        # value used to associate a Client session with a token
               ...
             }
           },
           ...
         }

    JSON Web Token (JWT) is a compact, URL-safe means of representing claims to be transferred between
    two parties. It is fully described in the RFC 7519, issued by the Internet Engineering Task Force
    (see https://www.rfc-editor.org/rfc/rfc7519.html).
    In this context, claims are pieces of information a token bears, and herein are loosely classified
    as token-related and account-related. All times are UTC.

    Token-related claims are mostly required claims, and convey information about the token itself:
       "exp": <timestamp>        # expiration time
       "iat": <timestamp>        # issued at
       "iss": <string>           # issuer (for remote providers, URL to obtain and validate the access tokens)
       "jti": <string>           # JWT id
       "sub": <string>           # subject (the account identification)
       "nat": <string>           # nature of token (A: access; R: refresh) - locally issued tokens, only
       # optional:
       "aud": <string>           # token audience
       "nbt": <timestamp>        # not before time

    Account-related claims are optional claims, and convey information about the registered account they belong to.
    Alhough they can be freely specified, these are some of the most commonly used claims:
       "birthdate": <string>     # subject's birth date
       "email": <string>         # subject's email
       "gender": <string>        # subject's gender
       "name": <string>          # subject's name
       "roles": <List[str]>      # subject roles
       "nonce": <string>         # value used to associate a client session with a token
    """
    def __init__(self) -> None:
        """
        Initizalize the token access data.
        """
        self.access_lock: Lock = Lock()
        self.access_data: dict[str, Any] = {}

    def add_access(self,
                   account_id: str,
                   reference_url: str,
                   claims: dict[str, Any],
                   access_max_age: int,
                   refresh_max_age: int,
                   grace_interval: int,
                   token_audience: str,
                   token_nonce: str,
                   request_timeout: int,
                   remote_provider: bool,
                   logger: Logger = None) -> None:
        """
        Add to storage the parameters needed to produce and validate JWT tokens for *account_id*.

        The parameter *claims* may contain account-related claims, only. Ideally, it should contain,
        at a minimum, "birthdate", "email", "gender", "name", and "roles".
        If the token provider is local, then the token-related claims are created at token issuing time.
        If the token provider is remote, all claims are sent to it at token request time.

        :param account_id: the account identification
        :param reference_url: the reference URL (for remote providers, URL to obtain and validate the JWT tokens)
        :param claims: the JWT claimset, as key-value pairs
        :param access_max_age: access token duration, in seconds
        :param refresh_max_age: refresh token duration, in seconds
        :param grace_interval: time to wait for token to be valid, in seconds
        :param token_audience: the audience the token is intended for
        :param token_nonce: optional value used to associate a client session with a token
        :param request_timeout: timeout for the requests to the reference URL
        :param remote_provider: whether the JWT provider is a remote server
        :param logger: optional logger
        """
        # build and store the access data for the account
        with self.access_lock:
            if account_id not in self.access_data:
                self.access_data[account_id] = {
                    "reference_url": reference_url,
                    "access-max-age": access_max_age,
                    "refresh-max-age": refresh_max_age,
                    "grace-interval": grace_interval,
                    "token-audience": token_audience,
                    "token-nonce": token_nonce,
                    "request-timeout": request_timeout,
                    "remote-provider": remote_provider,
                    "claims": claims or {}
                }
                if logger:
                    logger.debug(f"JWT data added for '{account_id}'")
            elif logger:
                logger.warning(f"JWT data already exists for '{account_id}'")

    def remove_access(self,
                      account_id: str,
                      logger: Logger) -> bool:
        """
        Remove from storage the access data for *account_id*.

        :param account_id: the account identification
        :param logger: optional logger
        return: *True* if the access data was removed, *False* otherwise
        """
        account_data: dict[str, Any] | None
        with self.access_lock:
            account_data = self.access_data.pop(account_id, None)

        if logger:
            if account_data:
                logger.debug(f"Removed JWT data for '{account_id}'")
            else:
                logger.warning(f"No JWT data found for '{account_id}'")

        return account_data is not None

    def issue_tokens(self,
                     account_id: str,
                     account_claims: dict[str, Any] = None,
                     logger: Logger = None) -> dict[str, Any]:
        """
        Issue and return the JWT access and refresh tokens for *account_id*.

        Structure of the return data:
        {
          "access_token": <jwt-token>,
          "created_in": <timestamp>,
          "expires_in": <seconds-to-expiration>,
          "refresh_token": <jwt-token>
        }

        :param account_id: the account identification
        :param account_claims: if provided, may supercede registered account-related claims
        :param logger: optional logger
        :return: the JWT token data, or *None* if error
        :raises InvalidTokenError: token is invalid
        :raises InvalidKeyError: authentication key is not in the proper format
        :raises ExpiredSignatureError: token and refresh period have expired
        :raises InvalidSignatureError: signature does not match the one provided as part of the token
        :raises ImmatureSignatureError: 'nbf' or 'iat' claim represents a timestamp in the future
        :raises InvalidAudienceError: 'aud' claim does not match one of the expected audience
        :raises InvalidAlgorithmError: the specified algorithm is not recognized
        :raises InvalidIssuerError: 'iss' claim does not match the expected issuer
        :raises InvalidIssuedAtError: 'iat' claim is non-numeric
        :raises MissingRequiredClaimError: a required claim is not contained in the claimset
        :raises RuntimeError: error accessing the revocation database, or
                              the remote JWT provider failed to return a token
        """
        # initialize the return variable
        result: dict[str, Any] | None = None

        # process the data in storage
        with (self.access_lock):
            account_data: dict[str, Any] = self.access_data.get(account_id)

            # was the JWT data obtained ?
            if account_data:
                # yes, proceed
                errors: list[str] = []
                current_claims: dict[str, Any] = account_data.get("claims").copy()
                if account_claims:
                    current_claims.update(account_claims)

                # obtain new tokens
                current_claims["jti"] = str_random(size=32,
                                                   chars=string.ascii_letters + string.digits)
                current_claims["iss"] = account_data.get("reference-url")

                # where is the JWT service provider ?
                if account_data.get("remote-provider"):
                    # JWT service is being provided by a remote server
                    # Structure of the return data:
                    # {
                    #   "access_token": <jwt-token>,
                    #   "created_in": <timestamp>,
                    #   "expires_in": <seconds-to-expiration>,
                    #   "refresh_token": <jwt-token>
                    #   ...
                    # }
                    result = _jwt_request_token(errors=errors,
                                                reference_url=current_claims.get("iss"),
                                                claims=current_claims,
                                                timeout=account_data.get("request-timeout"),
                                                logger=logger)
                    if errors:
                        raise RuntimeError(" - ".join(errors))
                else:
                    # JWT service is being provided locally
                    just_now: int = int(datetime.now(tz=timezone.utc).timestamp())
                    current_claims["iat"] = just_now

                    # retrieve the refresh token associated with the account id
                    refresh_token: str | None = None
                    if JWT_DB_ENGINE:
                        from pypomes_db import db_select, db_delete
                        if JWT_ROTATE_TOKENS:
                            db_delete(errors=errors,
                                      delete_stmt=f"DELETE FROM {JWT_DB_TABLE} "
                                                  f"WHERE account_id = '{account_id}'",
                                      logger=logger)
                        else:
                            recs: list[tuple[str]] = db_select(errors=errors,
                                                               sel_stmt=f"SELECT jwt_token FROM {JWT_DB_TABLE} "
                                                                        f"WHERE account_id = '{account_id}'",
                                                               max_count=1,
                                                               logger=logger)
                            if recs:
                                refresh_token = recs[0][0]
                        if errors:
                            raise RuntimeError(" - ".join(errors))

                    # was it obtained ?
                    if not refresh_token:
                        # no, issue a new one
                        current_claims["exp"] = just_now + account_data.get("refresh-max-age")
                        current_claims["nat"] = "R"
                        # may raise an exception
                        refresh_token: str = jwt.encode(payload=current_claims,
                                                        key=JWT_ENCODING_KEY,
                                                        algorithm=JWT_DEFAULT_ALGORITHM)
                        # persist the new refresh token
                        if JWT_DB_ENGINE:
                            from pypomes_db import db_insert
                            db_insert(errors=errors,
                                      insert_stmt=f"INSERT INTO {JWT_DB_TABLE}",
                                      insert_data={"account_id": account_id,
                                                   "jwt_token": refresh_token},
                                      logger=logger)
                            if errors:
                                raise RuntimeError(" - ".join(errors))

                    # issue the access token
                    current_claims["nat"] = "A"
                    current_claims["exp"] = just_now + account_data.get("access-max-age")
                    # may raise an exception
                    access_token: str = jwt.encode(payload=current_claims,
                                                   key=JWT_ENCODING_KEY,
                                                   algorithm=JWT_DEFAULT_ALGORITHM)
                    # return the token data
                    result = {
                        "access_token": access_token,
                        "created_in": current_claims.get("iat"),
                        "expires_in": current_claims.get("exp"),
                        "refresh_token": refresh_token
                    }
            else:
                # JWT access data not found
                err_msg: str = f"No JWT access data found for '{account_id}'"
                if logger:
                    logger.error(err_msg)
                raise RuntimeError(err_msg)

        return result


def _jwt_request_token(errors: list[str],
                       reference_url: str,
                       claims: dict[str, Any],
                       timeout: int = None,
                       logger: Logger = None) -> dict[str, Any]:
    """
    Obtain and return the JWT token from *reference_url*, along with its duration.

    Expected structure of the return data:
    {
      "access_token": <jwt-token>,
      "expires_in": <seconds-to-expiration>
    }
    It is up to the invoker to make sure that the *claims* data conform to the requirements
    of the provider issuing the JWT token.

    :param errors: incidental errors
    :param reference_url: the reference URL for obtaining JWT tokens
    :param claims: the JWT claimset, as expected by the issuing server
    :param timeout: request timeout, in seconds (defaults to *None*)
    :param logger: optional logger
    """
    # initialize the return variable
    result: dict[str, Any] | None = None

    # request the JWT token
    if logger:
        logger.debug(f"POST request JWT token to '{reference_url}'")
    response: Response = requests.post(
        url=reference_url,
        json=claims,
        timeout=timeout
    )

    # was the request successful ?
    if response.status_code in [200, 201, 202]:
        # yes, save the access token data returned
        result = response.json()
        if logger:
            logger.debug(f"JWT token obtained: {result}")
    else:
        # no, report the problem
        err_msg: str = f"POST request to '{reference_url}' failed: {response.reason}"
        if response.text:
            err_msg += f" - {response.text}"
        if logger:
            logger.error(err_msg)
        errors.append(err_msg)

    return result



import jwt
from flask import Request, Response, request
from logging import Logger
from typing import Any, Literal

from .jwt_constants import (
    JWT_ACCESS_MAX_AGE, JWT_REFRESH_MAX_AGE,
    JWT_DEFAULT_ALGORITHM, JWT_DECODING_KEY,
    JWT_DB_ENGINE, JWT_DB_TABLE
)
from .jwt_data import JwtData

# the JWT data object
__jwt_data: JwtData = JwtData()


def jwt_needed(func: callable) -> callable:
    """
    Create a decorator to authenticate service endpoints with JWT tokens.

    :param func: the function being decorated
    """
    # ruff: noqa: ANN003
    def wrapper(*args, **kwargs) -> Response:
        response: Response = jwt_verify_request(request=request)
        return response if response else func(*args, **kwargs)

    # prevent a rogue error ("View function mapping is overwriting an existing endpoint function")
    wrapper.__name__ = func.__name__

    return wrapper


def jwt_assert_access(account_id: str) -> bool:
    """
    Determine whether access for *account_id* has been established.

    :param account_id: the account identification
    :return: *True* if access data exists for *account_id*, *False* otherwise
    """
    return __jwt_data.access_data.get(account_id) is not None


def jwt_set_access(account_id: str,
                   reference_url: str,
                   claims: dict[str, Any],
                   access_max_age: int = JWT_ACCESS_MAX_AGE,
                   refresh_max_age: int = JWT_REFRESH_MAX_AGE,
                   grace_interval: int = None,
                   token_audience: str = None,
                   token_nonce: str = None,
                   request_timeout: int = None,
                   remote_provider: bool = True,
                   logger: Logger = None) -> None:
    """
    Set the data needed to obtain JWT tokens for *account_id*.

    :param account_id: the account identification
    :param reference_url: the reference URL (for remote providers, URL to obtain and validate the JWT tokens)
    :param claims: the JWT claimset, as key-value pairs
    :param access_max_age: access token duration, in seconds
    :param refresh_max_age: refresh token duration, in seconds
    :param grace_interval: optional time to wait for token to be valid, in seconds
    :param token_audience: optional audience the token is intended for
    :param token_nonce: optional value used to associate a client session with a token
    :param request_timeout: timeout for the requests to the reference URL
    :param remote_provider: whether the JWT provider is a remote server
    :param logger: optional logger
    """
    if logger:
        logger.debug(msg=f"Register access data for '{account_id}'")

    # extract the claims provided in the reference URL's query string
    pos: int = reference_url.find("?")
    if pos > 0:
        params: list[str] = reference_url[pos+1:].split(sep="&")
        for param in params:
            claims[param.split("=")[0]] = param.split("=")[1]
        reference_url = reference_url[:pos]

    # register the JWT service
    __jwt_data.add_access(account_id=account_id,
                          reference_url=reference_url,
                          claims=claims,
                          access_max_age=access_max_age,
                          refresh_max_age=refresh_max_age,
                          grace_interval=grace_interval,
                          token_audience=token_audience,
                          token_nonce=token_nonce,
                          request_timeout=request_timeout,
                          remote_provider=remote_provider,
                          logger=logger)


def jwt_remove_access(account_id: str,
                      logger: Logger = None) -> bool:
    """
    Remove from storage the JWT access data for *account_id*.

    :param account_id: the account identification
    :param logger: optional logger
    return: *True* if the access data was removed, *False* otherwise
    """
    if logger:
        logger.debug(msg=f"Remove access data for '{account_id}'")

    return __jwt_data.remove_access(account_id=account_id,
                                    logger=logger)


def jwt_validate_token(errors: list[str] | None,
                       token: str,
                       nature: Literal["A", "R"] = None,
                       logger: Logger = None) -> bool:
    """
    Verify if *token* ia a valid JWT token.

    Raise an appropriate exception if validation failed.

    :param errors: incidental error messages
    :param token: the token to be validated
    :param nature: optionally validate the token's nature ("A": access token, "R": refresh token)
    :param logger: optional logger
    :return: *True* if token is valid, *False* otherwise
    """
    if logger:
        logger.debug(msg=f"Validate JWT token '{token}'")

    err_msg: str | None = None
    try:
        # raises:
        #   InvalidTokenError: token is invalid
        #   InvalidKeyError: authentication key is not in the proper format
        #   ExpiredSignatureError: token and refresh period have expired
        #   InvalidSignatureError: signature does not match the one provided as part of the token
        claims: dict[str, Any] = jwt.decode(jwt=token,
                                            key=JWT_DECODING_KEY,
                                            algorithms=[JWT_DEFAULT_ALGORITHM])
        if nature and "nat" in claims and nature != claims.get("nat"):
            nat: str = "an access" if nature == "A" else "a refresh"
            err_msg = f"Token is not {nat} token"
    except Exception as e:
        err_msg = str(e)

    if err_msg:
        if logger:
            logger.error(msg=err_msg)
        if isinstance(errors, list):
            errors.append(err_msg)
    elif logger:
        logger.debug(msg=f"Token '{token}' is valid")

    return err_msg is None


def jwt_revoke_tokens(errors: list[str] | None,
                      account_id: str,
                      logger: Logger = None) -> bool:
    """
    Revoke all refresh tokens associated with *account_id*.

    Revoke operations require access to a database table defined by *JWT_DB_TABLE*.

    :param errors: incidental error messages
    :param account_id: the account identification
    :param logger: optional logger
    :return: *True* if operation could be performed, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    if logger:
        logger.debug(msg=f"Revoking refresh tokens of '{account_id}'")

    op_errors: list[str] = []
    if JWT_DB_ENGINE:
        from pypomes_db import db_delete
        delete_stmt: str = (f"DELETE FROM {JWT_DB_TABLE} "
                            f"WHERE account_id = '{account_id}'")
        db_delete(errors=op_errors,
                  delete_stmt=delete_stmt,
                  logger=logger)
    else:
        op_errors.append("Database access for token revocation has not been specified")

    if op_errors:
        if logger:
            logger.error(msg="; ".join(op_errors))
        if isinstance(errors, list):
            errors.extend(op_errors)
    else:
        result = True

    return result


def jwt_get_tokens(errors: list[str] | None,
                   account_id: str,
                   account_claims: dict[str, Any] = None,
                   refresh_token: str = None,
                   logger: Logger = None) -> dict[str, Any]:
    """
    Issue or refresh, and return, the JWT token data associated with *account_id*.

    If *refresh_token* is provided, its claims are used on issuing the new tokens,
    and claims in *account_claims*, if any, are ignored.

    Structure of the return data:
    {
      "access_token": <jwt-token>,
      "created_in": <timestamp>,
      "expires_in": <seconds-to-expiration>,
      "refresh_token": <jwt-token>
    }

    :param errors: incidental error messages
    :param account_id: the account identification
    :param account_claims: if provided, may supercede registered custom claims
    :param refresh_token: if provided, defines a token refresh operation
    :param logger: optional logger
    :return: the JWT token data, or *None* if error
    """
    # inicialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg=f"Retrieve JWT token data for '{account_id}'")
    op_errors: list[str] = []
    if refresh_token:
        account_claims = jwt_get_claims(errors=op_errors,
                                        token=refresh_token)
        if not op_errors and account_claims.get("nat") != "R":
            op_errors.extend("Invalid parameters")

    if not op_errors:
        try:
            result = __jwt_data.issue_tokens(account_id=account_id,
                                             account_claims=account_claims)
            if logger:
                logger.debug(msg=f"Data is '{result}'")
        except Exception as e:
            # token issuing failed
            op_errors.append(str(e))

    if op_errors:
        if logger:
            logger.error("; ".join(op_errors))
        if isinstance(errors, list):
            errors.extend(op_errors)

    return result


def jwt_get_claims(errors: list[str] | None,
                   token: str,
                   logger: Logger = None) -> dict[str, Any]:
    """
    Obtain and return the claims set of a JWT *token*.

    :param errors: incidental error messages
    :param token: the token to be inspected for claims
    :param logger: optional logger
    :return: the token's claimset, or *None* if error
    """
    # initialize the return variable
    result: dict[str, Any] | None = None

    if logger:
        logger.debug(msg=f"Retrieve claims for token '{token}'")

    try:
        claims: dict[str, Any] = jwt.decode(jwt=token,
                                            options={"verify_signature": False})
        if claims.get("nat") in ["A", "R"]:
            result = jwt.decode(jwt=token,
                                key=JWT_DECODING_KEY,
                                algorithms=[JWT_DEFAULT_ALGORITHM])
        else:
            result = claims
    except Exception as e:
        if logger:
            logger.error(msg=str(e))
        if isinstance(errors, list):
            errors.append(str(e))

    return result


def jwt_verify_request(request: Request,
                       logger: Logger = None) -> Response:
    """
    Verify wheher the HTTP *request* has the proper authorization, as per the JWT standard.

    :param request: the request to be verified
    :param logger: optional logger
    :return: *None* if the request is valid, otherwise a *Response* object reporting the error
    """
    # initialize the return variable
    result: Response | None = None

    if logger:
        logger.debug(msg="Validate a JWT token")
    err_msg: str | None = None

    # retrieve the authorization from the request header
    auth_header: str = request.headers.get("Authorization")

    # was a 'Bearer' authorization obtained ?
    if auth_header and auth_header.startswith("Bearer "):
        # yes, extract and validate the JWT access token
        token: str = auth_header.split(" ")[1]
        if logger:
            logger.debug(msg=f"Token is '{token}'")
        errors: list[str] = []
        jwt_validate_token(errors=errors,
                           nature="A",
                           token=token)
        if errors:
            err_msg = "; ".join(errors)
    else:
        # no 'Bearer' found, report the error
        err_msg = "Request header has no 'Bearer' data"

    # log the error and deny the authorization
    if err_msg:
        if logger:
            logger.error(msg=err_msg)
        result = Response(response="Authorization failed",
                          status=401)

    return result
